@@ -2,9 +2,9 @@
 #this repository contains the full copyright notices and license terms.
 from decimal import Decimal
 from trytond.model import Workflow, ModelView, ModelSQL, fields
-from trytond.pyson import Eval, If
+from trytond.pyson import Eval, If, Bool
 from trytond.transaction import Transaction
-from trytond.backend import TableHandler
+from trytond import backend
 from trytond.pool import Pool
 
 __all__ = ['Statement', 'Line']
@@ -19,21 +19,20 @@ class Statement(Workflow, ModelSQL, ModelView):
     company = fields.Many2One('company.company', 'Company', required=True,
         select=True, states=_STATES, domain=[
             ('id', If(Eval('context', {}).contains('company'), '=', '!='),
-                Eval('context', {}).get('company', 0)),
+                Eval('context', {}).get('company', -1)),
             ],
         depends=_DEPENDS)
     journal = fields.Many2One('account.statement.journal', 'Journal',
-        required=True,
+        required=True, select=True,
         domain=[
-            ('company', '=', Eval('context', {}).get('company', 0)),
+            ('company', '=', Eval('context', {}).get('company', -1)),
             ],
         states={
-            'readonly': (Eval('state') != 'draft') | Eval('lines'),
+            'readonly': (Eval('state') != 'draft') | Eval('lines', [0]),
             },
-        on_change=['journal', 'state', 'lines'], select=True,
-        depends=['state', 'lines'])
-    currency_digits = fields.Function(fields.Integer('Currency Digits',
-            on_change_with=['journal']), 'on_change_with_currency_digits')
+        depends=['state'])
+    currency_digits = fields.Function(fields.Integer('Currency Digits'),
+        'on_change_with_currency_digits')
     date = fields.Date('Date', required=True, states=_STATES, depends=_DEPENDS,
         select=True)
     start_balance = fields.Numeric('Start Balance', required=True,
@@ -45,13 +44,12 @@ class Statement(Workflow, ModelSQL, ModelView):
     balance = fields.Function(
         fields.Numeric('Balance',
             digits=(16, Eval('currency_digits', 2)),
-            on_change_with=['start_balance', 'end_balance'],
             depends=['currency_digits']),
         'on_change_with_balance')
     lines = fields.One2Many('account.statement.line', 'statement',
         'Transactions', states={
             'readonly': (Eval('state') != 'draft') | ~Eval('journal'),
-            }, on_change=['lines', 'journal'],
+            },
         depends=['state', 'journal'])
     state = fields.Selection([
         ('draft', 'Draft'),
@@ -59,8 +57,6 @@ class Statement(Workflow, ModelSQL, ModelView):
         ('cancel', 'Canceled'),
         ('posted', 'Posted'),
         ], 'State', readonly=True, select=True)
-    move_lines = fields.Function(fields.One2Many('account.move.line',
-        None, 'Move Lines'), 'get_move_lines')
 
     @classmethod
     def __setup__(cls):
@@ -70,9 +66,12 @@ class Statement(Workflow, ModelSQL, ModelView):
                 'wrong_end_balance': 'End Balance must be "%s".',
                 'delete_cancel': ('Statement "%s" must be cancelled before '
                     'deletion.'),
+                'paid_invoice_draft_statement': ('There are paid invoices on '
+                    'draft statements, do you want to proceed?'),
                 })
         cls._transitions |= set((
                 ('draft', 'validated'),
+                ('draft', 'cancel'),
                 ('validated', 'posted'),
                 ('validated', 'cancel'),
                 ('cancel', 'draft'),
@@ -88,12 +87,13 @@ class Statement(Workflow, ModelSQL, ModelView):
                     'invisible': Eval('state') != 'validated',
                     },
                 'cancel': {
-                    'invisible': Eval('state') != 'validated',
+                    'invisible': ~Eval('state').in_(['draft', 'validated']),
                     },
                 })
 
     @classmethod
     def __register__(cls, module_name):
+        TableHandler = backend.get('TableHandler')
         cursor = Transaction().cursor
 
         # Migration from 1.8: new field company
@@ -138,6 +138,7 @@ class Statement(Workflow, ModelSQL, ModelView):
             return company.currency.digits
         return 2
 
+    @fields.depends('journal', 'state', 'lines')
     def on_change_journal(self):
         res = {}
         if not self.journal:
@@ -155,6 +156,7 @@ class Statement(Workflow, ModelSQL, ModelView):
         res['start_balance'] = statement.end_balance
         return res
 
+    @fields.depends('journal')
     def on_change_with_currency_digits(self, name=None):
         if self.journal:
             return self.journal.currency.digits
@@ -180,36 +182,20 @@ class Statement(Workflow, ModelSQL, ModelView):
 
     @classmethod
     def search_rec_name(cls, name, clause):
-        statements = cls.search(['OR',
-                ('start_balance',) + tuple(clause[1:]),
-                ('end_balance',) + tuple(clause[1:]),
-                ])
-        if statements:
-            return [('id', 'in', [s.id for s in statements])]
         return [('journal',) + tuple(clause[1:])]
 
-    def get_move_lines(self, name):
-        '''
-        Return the move lines that have been generated by the statement.
-        '''
-        move_lines = []
-        for line in self.lines:
-            if not line.move:
-                continue
-            for move_line in line.move.lines:
-                move_lines.append(move_line.id)
-        return move_lines
-
     def get_end_balance(self, name):
         end_balance = self.start_balance
         for line in self.lines:
             end_balance += line.amount
         return end_balance
 
+    @fields.depends('start_balance', 'end_balance')
     def on_change_with_balance(self, name=None):
         return ((getattr(self, 'end_balance', 0) or 0)
             - (getattr(self, 'start_balance', 0) or 0))
 
+    @fields.depends('lines', 'journal')
     def on_change_lines(self):
         pool = Pool()
         Currency = pool.get('currency.currency')
@@ -220,32 +206,34 @@ class Statement(Workflow, ModelSQL, ModelView):
         if self.journal and self.lines:
             invoices = set()
             for line in self.lines:
-                if line.invoice:
+                if getattr(line, 'invoice', None):
                     invoices.add(line.invoice)
             invoice_id2amount_to_pay = {}
             for invoice in invoices:
                 with Transaction().set_context(date=invoice.currency_date):
-                    invoice_id2amount_to_pay[invoice.id] = (
+                    if invoice.type in ('out_invoice', 'in_credit_note'):
+                        sign = -1
+                    else:
+                        sign = 1
+                    invoice_id2amount_to_pay[invoice.id] = sign * (
                         Currency.compute(invoice.currency,
                             invoice.amount_to_pay, self.journal.currency))
 
-            for line in self.lines or []:
-                if line.invoice and line.id:
+            line_offset = 0
+            for index, line in enumerate(self.lines or []):
+                if getattr(line, 'invoice', None) and line.id:
                     amount_to_pay = invoice_id2amount_to_pay[line.invoice.id]
-                    if abs(line.amount) > amount_to_pay:
-                        res['lines'].setdefault('update', [])
-                        if self.journal.currency.is_zero(amount_to_pay):
-                            res['lines']['update'].append({
+                    res['lines'].setdefault('update', [])
+                    if (not self.journal.currency.is_zero(amount_to_pay)
+                            and getattr(line, 'amount', None)
+                            and (line.amount >= 0) == (amount_to_pay <= 0)):
+                        res['lines']['update'].append({
                                 'id': line.id,
-                                'invoice': None,
-                                })
-                        else:
-                            res['lines']['update'].append({
-                                'id': line.id,
-                                'amount': (amount_to_pay
-                                        if line.amount >= 0
-                                        else -amount_to_pay),
+                                'amount': (amount_to_pay.copy_sign(line.amount)
+                                    if abs(line.amount) > abs(amount_to_pay)
+                                    else line.amount)
                                 })
+                        if abs(line.amount) > abs(amount_to_pay):
                             res['lines'].setdefault('add', [])
                             vals = {}
                             for field_name, field in Line._fields.iteritems():
@@ -256,20 +244,26 @@ class Statement(Workflow, ModelSQL, ModelView):
                                 if (value and field._type in ('many2one',
                                             'one2one')):
                                     vals[field_name] = value.id
-                                    vals[field_name + '.rec_name'] = \
-                                        value.rec_name
+                                    if value.id >= 0:
+                                        vals[field_name + '.rec_name'] = \
+                                            value.rec_name
                                 else:
                                     vals[field_name] = value
                             del vals['id']
-                            vals['amount'] = (abs(line.amount)
-                                - amount_to_pay)
-                            if line.amount < 0:
-                                vals['amount'] = - vals['amount']
+                            vals['amount'] = line.amount + amount_to_pay
                             vals['invoice'] = None
                             del vals['invoice.rec_name']
-                            res['lines']['add'].append(vals)
-                    invoice_id2amount_to_pay[line.invoice.id] = \
-                        amount_to_pay - abs(line.amount)
+                            line_offset += 1
+                            res['lines']['add'].append((index + line_offset, vals))
+                            invoice_id2amount_to_pay[line.invoice.id] = 0
+                        else:
+                            invoice_id2amount_to_pay[line.invoice.id] = (
+                                line.amount + amount_to_pay)
+                    else:
+                        res['lines']['update'].append({
+                                'id': line.id,
+                                'invoice': None,
+                                })
         return res
 
     @classmethod
@@ -291,8 +285,11 @@ class Statement(Workflow, ModelSQL, ModelView):
     @ModelView.button
     @Workflow.transition('validated')
     def validate_statement(cls, statements):
-        Lang = Pool().get('ir.lang')
+        pool = Pool()
+        Lang = pool.get('ir.lang')
+        Line = pool.get('account.statement.line')
 
+        statement_lines = []
         for statement in statements:
             computed_end_balance = statement.start_balance
             for line in statement.lines:
@@ -309,6 +306,20 @@ class Statement(Workflow, ModelSQL, ModelView):
                     error_args=(amount,))
             for line in statement.lines:
                 line.create_move()
+                statement_lines.append(line.id)
+        cls.write(statements, {
+                'state': 'validated',
+                })
+        common_lines = Line.search([
+                ('statement.state', '=', 'draft'),
+                ('invoice.state', '=', 'paid'),
+                ])
+        if common_lines:
+            warning_key = '_'.join(str(l.id) for l in common_lines)
+            cls.raise_user_warning(warning_key, 'paid_invoice_draft_statement')
+            Line.write(common_lines, {
+                    'invoice': None,
+                    })
 
     @classmethod
     @ModelView.button
@@ -336,13 +347,10 @@ class Line(ModelSQL, ModelView):
             required=True, ondelete='CASCADE')
     date = fields.Date('Date', required=True)
     amount = fields.Numeric('Amount', required=True,
-        digits=(16, Eval('_parent_statement', {}).get('currency_digits', 2)),
-        on_change=['amount', 'party', 'account', 'invoice',
-            '_parent_statement.journal'])
-    party = fields.Many2One('party.party', 'Party',
-            on_change=['amount', 'party', 'invoice'])
+        digits=(16, Eval('_parent_statement', {}).get('currency_digits', 2)))
+    party = fields.Many2One('party.party', 'Party')
     account = fields.Many2One('account.account', 'Account', required=True,
-        on_change=['account', 'invoice'], domain=[
+        domain=[
             ('company', '=', Eval('_parent_statement', {}).get('company', 0)),
             ('kind', '!=', 'view'),
             ])
@@ -350,16 +358,13 @@ class Line(ModelSQL, ModelView):
     move = fields.Many2One('account.move', 'Account Move', readonly=True)
     invoice = fields.Many2One('account.invoice', 'Invoice',
         domain=[
-            ('party', '=', Eval('party')),
-            ('account', '=', Eval('account')),
+            If(Bool(Eval('party')), [('party', '=', Eval('party'))], []),
+            If(Bool(Eval('party')), [('account', '=', Eval('account'))], []),
             If(Eval('_parent_statement', {}).get('state') == 'draft',
                 ('state', '=', 'posted'),
                 ('state', '!=', '')),
             ],
-        states={
-            'readonly': (~Eval('amount') | ~Eval('party') | ~Eval('account')),
-            },
-        depends=['party', 'account', 'amount'])
+        depends=['party', 'account'])
 
     @classmethod
     def __setup__(cls):
@@ -382,6 +387,7 @@ class Line(ModelSQL, ModelView):
     def default_amount():
         return Decimal(0)
 
+    @fields.depends('amount', 'party', 'invoice')
     def on_change_party(self):
         res = {}
 
@@ -402,6 +408,8 @@ class Line(ModelSQL, ModelView):
                 res['invoice'] = None
         return res
 
+    @fields.depends('amount', 'party', 'account', 'invoice',
+        '_parent_statement.journal')
     def on_change_amount(self):
         Currency = Pool().get('currency.currency')
         res = {}
@@ -431,6 +439,7 @@ class Line(ModelSQL, ModelView):
                 res['invoice'] = None
         return res
 
+    @fields.depends('account', 'invoice')
     def on_change_account(self):
         res = {}
 
@@ -442,9 +451,25 @@ class Line(ModelSQL, ModelView):
                 res['invoice'] = None
         return res
 
+    @fields.depends('party', 'account', 'invoice')
+    def on_change_invoice(self):
+        changes = {}
+        if self.invoice:
+            if not self.party:
+                changes['party'] = self.invoice.party.id
+                changes['party.rec_name'] = self.invoice.party.rec_name
+            if not self.account:
+                changes['account'] = self.invoice.account.id
+                changes['account.rec_name'] = self.invoice.account.rec_name
+        return changes
+
     def get_rec_name(self, name):
         return self.statement.rec_name
 
+    @classmethod
+    def search_rec_name(cls, name, clause):
+        return [('statement.rec_name',) + tuple(clause[1:])]
+
     @classmethod
     def copy(cls, lines, default=None):
         if default is None:
@@ -490,7 +515,7 @@ class Line(ModelSQL, ModelView):
                 amount_to_pay = Currency.compute(self.invoice.currency,
                     self.invoice.amount_to_pay,
                     self.statement.journal.currency)
-            if amount_to_pay < abs(self.amount):
+            if abs(amount_to_pay) < abs(self.amount):
                 lang, = Lang.search([
                         ('code', '=', Transaction().language),
                         ])
@@ -506,7 +531,7 @@ class Line(ModelSQL, ModelView):
                     self.amount, self.statement.company.currency)
 
             reconcile_lines = self.invoice.get_reconcile_lines_for_amount(
-                abs(amount))
+                amount)
 
             for move_line in move.lines:
                 if move_line.account == self.invoice.account:
@@ -537,8 +562,9 @@ class Line(ModelSQL, ModelView):
         MoveLine = pool.get('account.move.line')
         Currency = Pool().get('currency.currency')
         zero = Decimal("0.0")
-        amount = Currency.compute(self.statement.journal.currency, self.amount,
-            self.statement.company.currency)
+        with Transaction().set_context(date=self.date):
+            amount = Currency.compute(self.statement.journal.currency,
+                self.amount, self.statement.company.currency)
         if self.statement.journal.currency != self.statement.company.currency:
             second_currency = self.statement.journal.currency.id
             amount_second_currency = abs(self.amount)
@@ -568,8 +594,8 @@ class Line(ModelSQL, ModelView):
         if self.account == account:
             self.raise_user_error('same_debit_credit_account', {
                     'account': self.account.rec_name,
-                    'line': self.account,
-                    'journal': self.journal,
+                    'line': self.rec_name,
+                    'journal': journal,
                     })
         move_lines.append(MoveLine(
                 description=self.description,
